import importlib.metadata

try:
    __version__ = importlib.metadata.version("datatailr")
except importlib.metadata.PackageNotFoundError:
    import toml  # type: ignore

    try:
        # load the version from pyproject.toml which is located two levels up relative to this file
        this_file_path = __file__
        pyproject_path = this_file_path.rsplit("/", 3)[0] + "/pyproject.toml"
        __version__ = toml.load(pyproject_path)["project"]["version"]
    except (FileNotFoundError, KeyError):
        __version__ = "unknown"



# *************************************************************************
#
#  Copyright (c) 2025 - Datatailr Inc.
#  All Rights Reserved.
#
#  This file is part of Datatailr and subject to the terms and conditions
#  defined in 'LICENSE.txt'. Unauthorized copying and/or distribution
#  of this file, in parts or full, via any medium is strictly prohibited.
# *************************************************************************

from __future__ import annotations

import json
from typing import List, Optional, Union

from datatailr import Group, User


class ACL:
    """
    A class to represent an Access Control List (ACL) for managing permissions.
    """

    def __init__(
        self,
        user: Union[User, str],
        group: Optional[Union[Group, str]] = None,
        permissions: Optional[List[str]] = None,
    ):
        if user is None:
            user = User.signed_user()
        self.user = user if isinstance(user, User) else User.get(user)
        if self.user is not None:
            self.group = (
                group if group and isinstance(group, Group) else self.user.primary_group
            )
        self.permissions = permissions or "rwr---"

        self.__group_can_read = False
        self.__group_can_write = False
        self.__user_can_read = False
        self.__user_can_write = False
        self.__world_readable = False

        self.__parse_permissions_string()

    def __parse_permissions_string(self):
        """
        Parse the permissions and set the corresponding flags.
        """
        if isinstance(self.permissions, str):
            self.permissions = list(self.permissions)
        if len(self.permissions) != 6:
            raise ValueError(
                "Permissions must be a list of 6 characters. representing 'r', 'w', and '-' for user, group, and world."
            )
        self.__user_can_read = self.permissions[0] == "r"
        self.__user_can_write = self.permissions[1] == "w"
        self.__group_can_read = self.permissions[2] == "r"
        self.__group_can_write = self.permissions[3] == "w"
        self.__world_readable = self.permissions[4] == "r"
        self.__world_writable = self.permissions[5] == "w"

    def _set_permissions_string(self):
        """
        Set the permissions string based on the current flags.
        """
        self.permissions = (
            f"{'r' if self.__user_can_read else '-'}"
            f"{'w' if self.__user_can_write else '-'}"
            f"{'r' if self.__group_can_read else '-'}"
            f"{'w' if self.__group_can_write else '-'}"
            f"{'r' if self.__world_readable else '-'}"
            f"{'w' if self.__world_writable else '-'}"
        )

    def __repr__(self):
        return (
            f"ACL(user={self.user}, group={self.group}, permissions={self.permissions})"
        )

    def to_dict(self):
        return {
            "user": self.user.name,
            "group": self.group.name if self.group else "dtusers",
            "group_can_read": self.__group_can_read,
            "group_can_write": self.__group_can_write,
            "user_can_read": self.__user_can_read,
            "user_can_write": self.__user_can_write,
            "world_readable": self.__world_readable,
            "world_writable": self.__world_writable,
        }

    @classmethod
    def from_dict(cls, acl_dict: dict) -> ACL:
        """
        Create an ACL instance from a dictionary.
        """
        user = User(acl_dict["user"])
        group = Group.get(acl_dict["group"]) if "group" in acl_dict else None
        acl = cls(user=user, group=group)
        acl.__group_can_read = acl_dict.get("group_can_read", False)
        acl.__group_can_write = acl_dict.get("group_can_write", False)
        acl.__user_can_read = acl_dict.get("user_can_read", False)
        acl.__user_can_write = acl_dict.get("user_can_write", False)
        acl.__world_readable = acl_dict.get("world_readable", False)
        acl.__world_writable = acl_dict.get("world_writable", False)
        acl._set_permissions_string()
        return acl

    def to_json(self):
        return json.dumps(self.to_dict())



##########################################################################
#
#  Copyright (c) 2025 - Datatailr Inc.
#  All Rights Reserved.
#
#  This file is part of Datatailr and subject to the terms and conditions
#  defined in 'LICENSE.txt'. Unauthorized copying and/or distribution
#  of this file, in parts or full, via any medium is strictly prohibited.
##########################################################################

"""
Module for caching arguments passed to batch jobs.

This module provides two backends for caching:
1. In-memory cache for local runs.
2. Persistent cache using the dt__Blob module for remote runs.

The cache stores arguments as a dictionary of dictionaries, where the outer dictionary's keys are job names
and the inner dictionaries contain the arguments.

This module is for internal use of the datatailr package.
"""

from datatailr.dt_json import json
import os
import pickle
from typing import Any, Dict, Optional

from datatailr import is_dt_installed, Blob
from datatailr.errors import DatatailrError


__BLOB_STORAGE__ = Blob()


class CacheNotFoundError(DatatailrError):
    """Custom error for cache operations."""

    def __init__(self, message: str):
        super().__init__(message)
        self.message = message


class ArgumentsCache:
    def __init__(self, use_persistent_cache: bool = is_dt_installed()):
        """
        Initialize the ArgumentsCache.

        :param use_persistent_cache: If True, use the persistent cache backend. Otherwise, use in-memory cache.
        """
        self.use_persistent_cache = use_persistent_cache
        if not self.use_persistent_cache:
            # Create a temp folder, for local caching
            os.makedirs("/tmp/datatailr/batch/arguments", exist_ok=True)
            os.makedirs("/tmp/datatailr/batch/results", exist_ok=True)

    def add_arguments(self, batch_id: str, arguments: Dict[str, Any]):
        """
        Add arguments to the cache for a specific job and batch run.

        :param batch_run_id: Identifier for the batch run.
        :param job_name: Name of the job.
        :param arguments: Dictionary of arguments to store.
        """
        path = f"/tmp/datatailr/batch/arguments/{batch_id}.pkl"
        if self.use_persistent_cache:
            self._add_to_persistent_cache(path, arguments)
        else:
            with open(path, "wb") as f:
                pickle.dump(arguments, f)

    def get_arguments(
        self, batch_id: str, job: str, batch_run_id: Optional[str]
    ) -> Dict[str, Any]:
        """
        Retrieve arguments from the cache for a specific job and batch run.

        :param batch_run_id: Identifier for the batch run.
        :param job_name: Name of the job.
        :return: Dictionary of arguments.
        """
        path = f"/tmp/datatailr/batch/arguments/{batch_id}.pkl"
        if self.use_persistent_cache and isinstance(job, str):
            try:
                arg_keys = self._get_from_persistent_cache(path)
            except RuntimeError:
                return {}
        else:
            if not os.path.exists(path):
                raise CacheNotFoundError(
                    f"Cache file not found: {path}. Ensure that the arguments have been cached."
                )
            with open(path, "rb") as f:
                try:
                    arg_keys = pickle.load(f)
                except EOFError:
                    return {}
                if not isinstance(arg_keys, dict):
                    raise TypeError(
                        f"Expected a dictionary for arguments, got {type(arg_keys)}"
                    )
        if batch_run_id is None:
            return arg_keys[job]
        args = {
            name: self.get_result(batch_run_id, value)
            for name, value in arg_keys[job].items()
        }
        return args

    def add_result(self, batch_run_id: str, job: str, result: Any):
        """
        Add the result of a batch job to the cache.

        :param batch_run_id: Identifier for the batch run.
        :param job: Name of the job.
        :param result: Result of the batch job.
        """
        path = f"/tmp/datatailr/batch/results/{batch_run_id}/{job}.pkl"
        if self.use_persistent_cache and isinstance(job, str):
            self._add_to_persistent_cache(path, result)
        else:
            os.makedirs(os.path.dirname(path), exist_ok=True)
            with open(path, "wb") as f:
                pickle.dump(result, f)

    def get_result(self, batch_run_id: str, job: Any) -> Any:
        """
        Retrieve the result of a batch job from the cache.

        :param batch_run_id: Identifier for the batch run.
        :param job: Name of the job.
        :return: Result of the batch job.
        """
        path = f"/tmp/datatailr/batch/results/{batch_run_id}/{job}.pkl"
        if self.use_persistent_cache and isinstance(job, str):
            return self._get_from_persistent_cache(path)
        else:
            if not os.path.exists(path):
                return job
            with open(path, "rb") as f:
                try:
                    return pickle.load(f)
                except EOFError:
                    return None

    def _add_to_persistent_cache(self, path: str, blob: Any):
        """
        Add arguments to the persistent cache.
        This method serializes the blob using pickle and stores it in the Blob storage.
        :param path: Path in the Blob storage where the blob will be stored.
        :param blob: The blob to store, typically a dictionary of arguments.
        :raises TypeError: If the blob cannot be pickled.

        """
        path = path.replace("/tmp/", "")
        __BLOB_STORAGE__.put_blob(path, json.dumps(blob))

    def _get_from_persistent_cache(self, path: str) -> Any:
        """
        Retrieve arguments from the persistent cache.

        :param path: Path in the Blob storage where the blob is stored.
        """
        path = path.replace("/tmp/", "")
        data = __BLOB_STORAGE__.get_blob(path)
        return json.loads(data)
